self.__precacheManifest = (self.__precacheManifest || []).concat([
  {
    "revision": "d52117df20953279ee158609f127fa64",
    "url": "./index.html"
  },
  {
    "revision": "af08ea6b6607896320f4",
    "url": "./static/js/2.2bc12ddb.chunk.js"
  },
  {
    "revision": "065af7f54f052065d2d7d684b53f2ae2",
    "url": "./static/js/2.2bc12ddb.chunk.js.LICENSE.txt"
  },
  {
    "revision": "0e04dd2981e94467c748",
    "url": "./static/js/main.a8b332f3.chunk.js"
  },
  {
    "revision": "58b1ca89039dae1abe7e",
    "url": "./static/js/runtime-main.41b280b5.js"
  }
]);