(this.webpackJsonpstreamlit_extras_cookie_manager=this.webpackJsonpstreamlit_extras_cookie_manager||[]).push([[0],{5:function(e,t,a){e.exports=a(6)},6:function(e,t,a){"use strict";a.r(t);var s=a(4),n=a(0),i=null,o=new s.a;parent.CookieManager=o,window.CookieManager=o;n.a.events.addEventListener(n.a.RENDER_EVENT,(function(e){var t=e.detail,a=t.args,s=(t.disabled,t.theme,a.method),r=a.options,m=void 0===r?{}:r,p=m.name,l=m.value,g=m.expires,u=m.sameSite,c=m.secure,d=m.path,h={name:p,value:l,expires:new Date(g)||new Date((new Date).getTime()+432e5),sameSite:u||"strict",secure:c||!0,path:d||"/"},v={path:h.path,sameSite:h.sameSite},w="set"===s?o.set(p,l,h)||!0:"get"===s?o.get(p):"getAll"===s?o.getAll():"delete"===s?o.remove(p,v)||!0:null;w&&JSON.stringify(i)!=JSON.stringify(w)&&(i=w,n.a.setComponentValue(w),n.a.setComponentReady()),n.a.setFrameHeight(0)})),n.a.setComponentReady(),n.a.setFrameHeight(0)}},[[5,1,2]]]);
//# sourceMappingURL=main.a8b332f3.chunk.js.map